#!/usr/bin/env python

import sys
import rospy
import moveit_commander
import moveit_msgs.msg
import geometry_msgs.msg
from math import pi
from copy import deepcopy
from std_msgs.msg import String
from moveit_commander.conversions import pose_to_list
import gcode2contour as gc


class TwinsController(object):
    """
    Object to control the two arm setup.
    Useful wrappers aroung the MoveIt Commander functions
    """

    def __init__(self):

        moveit_commander.roscpp_initialize(sys.argv)
        rospy.init_node('twins_py_controller', anonymous=True)

        self.robot = moveit_commander.RobotCommander()
        self.arm1 = moveit_commander.MoveGroupCommander("robot1")
        self.arm2 = moveit_commander.MoveGroupCommander("robot2")

        self.arms = {
                "arm1": self.arm1,
                1:      self.arm1,
                "arm2": self.arm2,
                2:      self.arm2
                }

        self.display_trajectory_publisher = rospy.Publisher(
                '/move_group/display_planned_path',
                moveit_msgs.msg.DisplayTrajectory,
                queue_size=20
            )

        # self.move_group.get_planning_frame()
        # self.move_group.get_end_effector_link()
        # self.robot.get_current_state()
        # self.scene = moveit_commander.PlanningSceneInterface()
            # use this to add objects to world


    def go_to_joint(self, joint_state, arm):
        """
        joint state is list of joint angles (len 6)
        arm can be 1, 2, 'arm1' or 'arm2'
        """

        self.arms[arm].go(joint_state, wait=True)
        self.arms[arm].stop() #ensure no residual movement

        current_joints = self.arms[arm].get_current_joint_values()
        return current_joints


    def go_to_pose(self, pose_goal, arm):
        """
        pose_goal as geometry_msg.msg.Pose() object
        arm can be 1, 2, 'arm1' or 'arm2'
        """
        self.arms[arm].set_pose_target(pose_goal)
        plan = self.arms[arm].go(wait=True)
        self.arms[arm].stop()
        self.arms[arm].clear_pose_targets() #good practice
        return 


    def plan_cartesian_path(self, arm, scale=1):
        pass
#        waypoints = []
#
#        wpose = self.arms[arm].get_current_pose().pose
#        wpose.position.z -= scale * 0.1    # First move up (z)
#        wpose.position.y += scale * 0.2    # and sideways (y)
#        waypoints.append(deepcopy(wpose))
#
#        wpose.position.x += scale * 0.1    # Second move forward/backwards in (x)
#        waypoints.append(deepcopy(wpose))
#
#        wpose.position.y -= scale * 0.1    # Third move sideways (y)
#        waypoints.append(deepcopy(wpose))
#
#        (plan, fraction) = self.arms[arm].compute_cartesian_path(
#                             waypoints,     # waypoints to follow
#                             0.01,                # eef_step
#                             0.0)                 # jump_threshold
#
#        return plan, fraction


    def plan_contour(self, c, arm):
        """
        Contour object from gcode2contour object
            specifying postions in global coordinates

        Returns plan, and frac (which describes how
            much of the plan could be planned)
        """
        #TODO convert the following to always use vertical pose
        # Longterm todo: perpendicular to pose 
        wpose = self.arms[arm].get_current_pose().pose

        waypoints = []
        scale = 0.001 # convert mm to m
        for p in c.pos:
            wpose.position.x = p[0]*scale
            wpose.position.y = p[1]*scale
            wpose.position.z = p[2]*scale
            waypoints.append(deepcopy(wpose))

        plan, frac = self.arms[arm].compute_cartesian_path(
                waypoints, 
                0.1, # eef_step
                0.0  # jump_threshold
            )

        return plan, frac


    def display_trajectory(self, plan):
        display_trajectory = moveit_msgs.msg.DisplayTrajectory()
        display_trajectory.trajectory_start = self.robot.get_current_state()
        display_trajectory.trajectory.append(plan)
        self.display_trajectory_publisher.publish(display_trajectory);
        return


    def execute_plan(self, plan, arm):
        self.arms[arm].execute(plan, wait=True)
        return
