"""
Useful functions
"""
import numpy as np
from numpy import pi
import csv
from copy import deepcopy
# pip install git+https://github.com/cambel/ur_ikfast@3c9c8b5a1400ea7946b609a4097adae9c8d3c7a8
from ur_ikfast import ur_kinematics


# ROS Libraries
import rospy
from trac_ik_python.trac_ik import IK
from std_msgs.msg import Bool
from trajectory_msgs.msg import (
    JointTrajectoryPoint,
    JointTrajectory
)


ur_default = ur_kinematics.URKinematics('ur10')

class JTrajectory(object):

    def __init__(self, pos, times, robot, contour=None):
        assert (len(pos)==len(times)), "positions and times must be same len"
        self.positions = pos
        self.time = times
        self.contour = contour
        self.robot = robot
        self.hash = hash((self.contour, tuple(self.positions[0])))


    def __hash__(self):
        """
        to use JTrajectory as keys in dicts
        """
        return self.hash


    def __eq__(self, other):
        return self.hash == other.hash


def joints_at_time(traj, t):
    """
    get the joint position at a time (relative to start of traj)
    """
    if t==0:
        return traj.positions[0]

    wp_i = bisect_left(traj.time, t)

    # linear interpolate between the waypoints
    j1 = np.array(traj.positions[wp_i-1])
    j2 = np.array(traj.positions[wp_i])
    t1 = traj.time[wp_i-1]
    t2 = traj.time[wp_i]

    return (t-t1)/(t2-t1)*(j2-j1) + j1


default_joint_names = ['shoulder_pan_joint',
 'shoulder_lift_joint',
 'elbow_joint',
 'wrist_1_joint',
 'wrist_2_joint',
 'wrist_3_joint']


def JTraj2ROS(jtraj,joint_names=default_joint_names):
    """
    convert JTrajectory object to ROS JointTrajectory
    """
    j = JointTrajectory()
    j.joint_names = joint_names

    for i in range(len(jtraj.positions)):
        if i>0 and jtraj.time[i-1] >= jtraj.time[i]:
            print("Caught Double time stamp")
            if not np.allclose(jtraj.positions[i-1],jtraj.positions[i],rtol=1e-3,atol=1e-2):
                print(np.array(jtraj.positions[i-1]) - jtraj.positions[i])
                raise Exception("Time stamps equal but position changed")
            # catch double time stamps
            continue
        jp = JointTrajectoryPoint()
        jp.positions = jtraj.positions[i]
        jp.time_from_start = rospy.Duration(jtraj.time[i])
        j.points.append(jp)

    return j


def calibration_correction(jtraj, robot_desc_param, max_attempts=50, save=False, max_diff=0.7):
    """
    modify joint angles from default ur10 ikfast to calibrated arm
    """
    urdf_str = rospy.get_param(robot_desc_param)
    ik_solver = IK("world", "tool0", urdf_string=urdf_str)
    if save:
        csvfile = open('trac.csv','ab')
        writer = csv.writer(csvfile)

    cal_jtraj = deepcopy(jtraj)
    cal_jtraj.positions = []
    cal_jtraj.time = []

    n_skipped = 0
    for i in range(len(jtraj.positions)):
        attempts = 0
        cal_dist = 10
        intended_pose = ur_default.forward(jtraj.positions[i])
        orig_angles = np.array(jtraj.positions[i])
        while cal_dist>0.2 and attempts < max_attempts:
            attempts += 1
            temp = np.array(ik_solver.get_ik(orig_angles, *intended_pose))
            if temp is None or None in temp:
                print("IK failed")
                continue
            cal_dist_new = np.linalg.norm(np.array(jtraj.positions[i]) - temp)
            if cal_dist_new < max_diff and cal_dist_new < cal_dist:
                cal_dist = cal_dist_new
                cal_jtraj.positions.append(list(temp))
                cal_jtraj.time.append(jtraj.time[i])
#            if attempts >1 and cal_dist>max_diff:
#                print(attempts, cal_dist, cal_dist_new)
        if save:
            writer.writerow([list(jtraj.positions[i]), cal_dist])

        if cal_dist>max_diff:
            # cal_jtraj.positions[i] = jtraj.positions[i]
            print("Failed to find a close solution for waypoint {}, cal_dist = {}".format(i, cal_dist))
            print("skipping waypoint")
            n_skipped += 1
            if n_skipped>3:
                raise Exception("Too many waypoints skipped")
            if i==len(jtraj.positions) or i==0:
                raise Exception("First and last waypoints in a trajectory cannot be skipped")
        else:
            n_skipped = 0
    if save:
        csvfile.close()
    return cal_jtraj


def calib_corr_single(j,robot_desc_param):
    """
    modify joint angles from default ur10 ikfast to calibrated arm
    """
    urdf_str = rospy.get_param(robot_desc_param)
    ik_solver = IK("world", "tool0", urdf_string=urdf_str)
    intended_pose = ur_default.forward(j)
    j_out = list(ik_solver.get_ik(j, *intended_pose))
    if sum((np.array(j) - np.array(j_out))**2)>0.1:
        raise Exception("trac-ik solution too far away from original joints")
    return j_out


def speed_multiplier(jtraj, multiplier):
    """
    used for slowing down contour execution for testing
    """
    if isinstance(jtraj, JTrajectory):
        jtraj.time = [t*multiplier for t in jtraj.time]
    elif isinstance(jtraj, JointTrajectory):
        for p in jtraj.points:
            p.time_from_start = p.time_from_start*multiplier
            p.velocities = []
            p.accelerations = []
            p.effort = []
    return jtraj
